###############################################################################
# Mark Cowley
# Peter Wills Bioinformatics Centre
# Garvan Institute of Medical Research
#
# This software and its documentation are copyright (2010-2014) by Mark
# Cowley. All rights are reserved.

# This software is supplied without any warranty or guaranteed support
# whatsoever. Garvan Institute can not be responsible for its
# use, misuse, or functionality.

#
# This GenePattern module will log transform a GCT file. 
# It is aware of negative values and provides strategies for 
# log transforming these data.
#

## Future features / todo list.
# offset (DONE)
# help with offset
#		write summary to new file? 
# 		how many pts less than 1 (DONE)
# 	auto - shifts everything to >=1 before log transform(DONE) 
# log 10, vst 

##BUGS
# auto fails on missing data(DONE)
# fudebug.message and warning doesn't seem to be printing

## Command line.
# <R2.5> <libdir>LogTransform.R main -g<input.file> -o<output.file> -v<offset.opt> -l<libdir>
#
###############################################################################

###############################################################################
# Version History
# 1:
#	- version 1.
# 2:
#	- removed offset.val (got this error: Error in eval.with.vis(expr, envir, enclos) : cannot change value of locked binding for 'offset')
# 3:
# - added common.pwbc.R
# 4: 
#	- Check input.file end in .gct, add offset.val
# 5:
#	- auto offset. cmd line offset.val is now refered to as offset.opt
# 6:	
#	- Summary
# 7:	
# 	- auto fail on missing data
#  	- Output formatting and headings for each step
# 8:
#   	- cmd line offset.opt now offset.value for better genepattern usability
#	  value is still referred to as offset.opt in parsing
# Public Beta (gpqa)
# 3:    
#	- no missing data point(s) message typo
###############################################################################


###############################################################################
# static/global variables.
DEBUG <<- FALSE 
libdir <<- ""           # The libdir where the R files & libraries live
input.file <<- ""	# gct file
output.file <<- ""	# name of gct output file.

offset.val <<- 0	# [optional] offset
AUTOOFFSET <<- FALSE	# auto offset, move all data

NSIGFIGS <<- 5		# num decimal places.
xdata <<- NULL		# will contain the gct files contents after we import it.

###############################################################################

###############################################################################
#
# program loader function. This should be called first by genepattern.
#
###############################################################################
main <- function(...) {
	#
	# 1. parse the command line arguments, and setup the global variables (see above)
	#
	if(DEBUG) cat("Parsing command line\n")
	parseCmdLine(..., verbose=DEBUG)
	
	#
	# 2. source the relevant R libraries
	#
	source(paste(libdir, "common.R", sep=''))
	source(paste(libdir, "common.pwbc.R", sep=''))
	
	#
	# 3. do analysis
	#
	debug.message("entering LogTransform.analysis().\n")
	LogTransform.analysis()
	
	#
	# 4. cleanup
	#
	cleanup()
}


###############################################################################
#
# perform a base 10 log transformation analysis on a gct file
#
###############################################################################
LogTransform.analysis <- function() {
	debug.message(paste("SEARCH PATHS:", searchpaths(), "\n"))

	###########################################################
	# import the gct expression data.
	#
	cat("### Importing Expression data from gct file.\n")
	xdata <<- NULL
	xdata <<- read.gct(input.file)

	###########################################################
	
	###########################################################
	# Print a summary of the data 
	#
	cat(sprintf("%s contains:\n", basename(input.file)))
	cat(genSummary.gct(xdata))

	###########################################################

	###########################################################
	# Add an offset 
    cat("### Applying offset...\n")
	offset.res <- xdata
    min.val <- min(offset.res$data, na.rm=TRUE)
	if (AUTOOFFSET) { 
		if (min.val < 1) {
			offset.val <<- abs(min.val) + 1
		} else {
			offset.val <<- 0 
		}
	} 
	cat(sprintf("Values were offset by %f\n", offset.val))
	offset.res$data <- offset.res$data + offset.val 

	###########################################################
	
	###########################################################
	# Truncate the data
	#
    cat("### Truncating data...\n")
	trunc.res <- offset.res 
	if(any(trunc.res$data < 1, na.rm=TRUE)) {
		cat("Truncating values < 1 to 1 ...\n")
		cat(sprintf("There were %d values < 1\n", sum(trunc.res$data < 1, na.rm=TRUE)))
		trunc.res$data[ trunc.res$data < 1 ] <- 1
	}
	else {
		cat("All data was positive. no truncation needed.\n")
	}
	###########################################################


	###########################################################
	# log-base-2 transform the trunc.data
	# make sure you're not logging data that's already logged
    cat("### Applying log2 transform...\n")
	if(all(trunc.res$data < 60)) {
		cat("WARNING: data may already be log transformed\n")
	}

	log2.data <- trunc.res
	log2.data$data <- log2(log2.data$data)
	log2.data$data <- signif(log2.data$data, NSIGFIGS)
	# Missing data points:
	# These data are expected to be represented as empty strings in the GCT file
	# ie. An occurence of a missing data point will result 
	# in two tab delimiters next to each other: \t\t
	# R properly handles these in read.gct and print.gct
	# In an R data structure they will appear as NA

	###########################################################
	# export a summary of the transformed gct	
	
	cat("Log transformed file contains:\n")
	cat(genSummary.gct(log2.data))

	###########################################################
	
	
	###########################################################
	# export gct file
	#
	write.gct(log2.data, output.file, check.file.extension=TRUE)
	###########################################################
	
}


#########################################################
#                                                       #
#     Standard functions                                #
#                                                       #
#########################################################

#########################################################
#
# Generate a summary string of data read from a GCT file
# data should be read using read.gct ie. of the format:
# list(row.descriptions=row.descriptions, data=data))
#
# Kevin Ying 
genSummary.gct<- function(x) {
	
	totalCount <- ncol(x$data) * nrow(x$data)
	nbelow1 <- sum(x$data < 1, na.rm=TRUE)
	nbelow0 <- sum(x$data < 0, na.rm=TRUE)
	min	<- min(x$data, na.rm=TRUE)
	max	<- max(x$data, na.rm=TRUE)
	naCount <- sum(is.na(x$data))

	header <- sprintf("   %d samples and %d genes\n", ncol(x$data), nrow(x$data))
	below1 <- sprintf("   %d (%.3f %%) out of %d are < 1\n", nbelow1, signif(100*nbelow1/totalCount, 3), totalCount)  
	below0 <- sprintf("   %d (%.3f %%) out of %d are < 0\n", nbelow0, signif(100*nbelow0/totalCount, 3), totalCount)  
	stats  <- sprintf("   Min. : %f\n    Max. : %f\n", min, max)
	if (naCount > 0) {
		missing <- sprintf("   There were %d missing data points\n", naCount)
	} else {
		missing <- "   There were no missing data points\n"
	}

	#return (summary(xdata$data))
	return (paste("", header, below1, below0, stats, missing))
}

#########################################################
#
# The last function to be called. can be used to cleanup/move files around.
# .... in practice this rarely does anything at all.
#
# Mark Cowley
cleanup <- function() {
	x <- TRUE # i.e. do nothing.
}

#########################################################
#
# Parse the "..." arguments passed into the R code by GP.
# The commandline looks like this:
# <R2.5> <libdir>LogTransform.R main -g<input.file> -o<output.file> -v<offset.opt> -l<libdir>
#
# Mark Cowley
parseCmdLine <- function(..., verbose=FALSE) {
	.trim <- function(x) {
		sub("^[ \t]*([^ \t]+.*[^ \t]+)[ \t]+$", "\\1", x, perl=TRUE)
	}
	# foreach arg, remove the leading flag, (eg -'x') and return a list with
	# flags as the list names, and the value as the list values.
	# can handle spaces between the flag and the argument, though i'd expect that
	# GP would treat '-x <myfile>' as 2 args: '-x' and '<myfile>', which would thus
	# fail in my code which expects a value after the flag for each argument.
	.parse.command.line <- function(args) {
		result <- list()
		for(i in 1:length(args)) {
			flag <- substring(args[[i]], 0, 2)
			value <- substring(args[[i]], 3, nchar(args[[i]]))
			value <- .trim(value) # trim leading and trailing whitespace.
			if(flag=='') {
				next
			}	
			result[[flag]] <- value
		}
		result
	}
	if(verbose) cat(sprintf("Found %d arguments.\n", length(list(...))))

	args <- .parse.command.line(list(...))
	if(verbose) cat(paste(c(names(args), "\n"), collapse=" "))
	if(verbose) str(args)
	
	# Parse the expected arguments.
	input.file			<<- args[["-g"]] # gct
	output.file			<<- args[["-o"]] # string
	offset.opt			<<- args[["-v"]] # A number of the string "auto" 
	libdir				<<- args[["-l"]] # ie <libdir>

	suppressWarnings(offset.val <<- as.numeric(offset.opt))
	if (is.na(offset.val) && offset.opt == "auto") {
		AUTOOFFSET <<- TRUE 
		cat("Using auto offset, all values will be shifted by 1 + minimum value\n")
	} else if (is.na(offset.val)) { 
		stop("Offset value not recognised, enter a number or \"auto\"\n")
	}

	# are there any unexpected arguments??
	expected.args <- c("-g", "-o", "-l", "-v")
	remaining.args <- setdiff(names(args), expected.args)
	if( length(remaining.args)>0 )
		stop("Unknown flags: ", paste(remaining.args, collapse=", "), .call=FALSE)
	# check for duplicate flags
	if( any(table(expected.args) > 1) ) {
		t <- table(expected.args)
		dupflags <- names(t)[which(t > 1)]
		stop( sprintf("You have duplicated these flags in your command line: %s\n", paste(dupflags, collapse=", ")) )
	}

	if( !file.exists(input.file) ) {
		stop("Can't read the input GCT file\n")
	}
	# check that the input.file ends in gct.
	ext <- regexpr(paste("gct", "$",sep=""), tolower(input.file))
	if(ext[[1]] == -1) {
		stop("The input file should end in .gct")	
	}

	if( !file.exists(libdir) ) {
		stop("Can't read <libdir>\n")
	}
	if(nchar(output.file)<4) {
		stop("output.file must be the name of a valid GCT file.\n")
	}
	if(verbose) cat("Finished checking arguments.\n")
}

