###############################################################################
# Package Installation Functions.
#
setLibPath <- function(libdir) {
	libPath <- libdir
	if(!file.exists(libdir) && substr(libdir, nchar(libdir), nchar(libdir))=="/") {
		# remove trailing /
		libPath <- substr(libdir, 0, nchar(libdir)-1)
	}
	.libPaths(libPath)
}


isWindows <- function() {
	Sys.info()[["sysname"]]=="Windows"
}


isMac <- function() {
	Sys.info()[["sysname"]]=="Darwin" 
}


is.package.installed <- function(libdir, pkg) {
	f <- paste(libdir, pkg, sep='')
	return(file.exists(f) && file.info(f)[["isdir"]])
}


# install.package <- function(dir, windows, mac, other) {
# 	if(isWindows()) {
# 		f <- paste(dir, windows, sep="")
# 		.install.windows(f)
# 	} else if(isMac()) {
# 		f <- paste(dir, mac, sep="")
# 		.install.mac(f)
# 	} else { # install from source
# 		f <- paste(dir, other, sep="")
# 		.install.unix(f)
# 	}
# }
# 
# 
# .install.windows <- function(pkg) {
# 	if(DEBUG) {
# 		info("Installing windows package ", pkg)
# 	}
# 	suppressMessages(install.packages(pkg, .libPaths()[1], repos=NULL))
# }
# 
# .install.mac <- function(pkg) {
# 	.install.unix(pkg)
# }
# 
# .install.unix <- function(pkg) {
# 	if(DEBUG) {
# 		info("Installing package ", pkg)
# 	}
# 	lib <- .libPaths()[1]
# 	# cmd <- paste(file.path(R.home(), "bin", "R"), "CMD INSTALL --with-package-versions")
# 	cmd <- paste(file.path(R.home(), "bin", "R"), "CMD INSTALL")
# 	cmd <- paste(cmd, "-l", lib)
# 	cmd <- paste(cmd, " '", pkg, "'", sep = "")
# 	status <- suppressMessages(system(cmd))
# 	if (status != 0) 
# 		cat("\tpackage installation failed\n")
# }

# Install a package just using its name (eg 'limma'), looking into a dir that
# may contain lots of different packages for different platforms, eg limma.zip, limma.tar.gz, ...
# and installs it to the specified libdir.
#
# Parameters:
#	pkgname: a character(1), eg "limma"
#	pkgdir: path to where the source/binary bundles live.
#	lib: the library location to install to. see install.packages
#	verbose: TRUE for noisy...
#
# Value:
#	TRUE or FALSE depending on whether relevant package files were found or not.
#
# Mark Cowley, 2010-02-03
#
install.package.localdir <- function(pkgname, pkgdir, lib.loc, verbose=TRUE) {
	if( length(pkgname) > 1 )
		stop("only one package at a time.\n")
	
	status <- FALSE
	if( is.package.installed(lib.loc, pkgname) ) {
		status <- TRUE
	}
	else {
		# suffix <- switch(Sys.info()[["sysname"]], Windows=".zip", Darwin=".tgz", "tar.gz")
		suffix <- switch(Sys.info()[["sysname"]], Windows=".zip", "tar.gz")
		pattern <- paste(pkgname, "_[0-9.-]+", suffix, sep="") # eg limma_<version>.tgz

		# newer versions have higher version numbers & are thus later in the listing...
		pkg.file <- rev(dir(pkgdir, pattern=pattern, full=TRUE))
		if( length(pkg.file) > 0 ) {
			status <- TRUE
			pkg.file <- pkg.file[1] # choose the newest
			if( verbose ) {
				debug.message( sprintf("Found the following candidates. Choosing the first: %s\n",
					paste(pkg.file, collapse=", ")) )
				install.packages(pkg.file, lib=lib.loc, repos=NULL)
			}
			else {
				suppressMessages(install.packages(pkg.file, lib=lib.loc, repos=NULL))
			}
			R.css <- file.path(lib.loc, "R.css")
			if( file.exists(R.css) ) unlink(R.css)
		}
		else {
			status <- FALSE
		}
	}
	
	invisible(status)
}
###############################################################################


# MJC 2010-01-08
# returns a vector of filenames
crossplatform.unzip <- function(zip.filename, dest=NULL, verbose=FALSE) {
	if(is.null(dest)) {
		dest = getwd()
	}
	
	if( !file.exists(zip.filename) ) {
		stop(gettextf("zipfile '%s' not found", zip.filename), domain = NA)
	}
	
	res <- NULL
	if( isWindows() ) {
		# I am sure that the utils::unzip should work fine on Windows, but this is here because
		# I haven't had a chance to test it yet... MJC, 2010-01-08
		# This code was called install.packages.R->zip.unpack
		if((unzip <- getOption("unzip")) != "internal") {
			res <- system(paste(unzip, "-oq", zip.filename, "-d", dest),
				   show = FALSE, invisible = TRUE)
		} else {
			res <- unzip(zip.filename, files=NULL, exdir=dest)
		}
	}
	else {
		res <- utils::unzip(zip.filename, exdir=dest)
		# unzip <- getOption("unzip")
		# system(paste(unzip, "-q", zip.filename, "-d", dest))
	}

	# if a set of local files are zipped using 'compress' from OSX finder, 
	# you can get special files included in the zip. these should be removed.
	mocosx.system.files <- grep("__MACOSX", res, value=TRUE)
	if( length(mocosx.system.files) > 1 ) {
		res <- setdiff(res, mocosx.system.files)
		
		macosx.dir <- unique(dirname(mocosx.system.files))
		unlink(macosx.dir, recursive=TRUE)
	}
	
	if( verbose )
		cat("Found these files in the zip archive:\n", paste(res, collapse="\n"), "\n", sep="")
	
	return( res )
}

# trim whitespace from start and end of strings.
trim <- function(s) {
	s <- sub('[ \t]+$', '', s, extended = TRUE) 
	s <- sub('^[ \t]+', '', s, extended = TRUE) 
	s
}


# Print a message, only if the global DEBUG flag is TRUE.
debug.message <- function(msg) {
	if(DEBUG)  {
		msg <- paste("DEBUG:", msg)
		cat(msg)
	}
}


cat.matrix <- function (x, pad=TRUE, sep=" ") {
	x <- rbind(colnames(x), x)
	x <- cbind(rownames(x), x)
	if( pad ) {
		for(i in 1:ncol(x)) {
			x[,i] <- as.character(x[,i])
			w <- max(nchar(x[,i]))
			cmd <- paste("%",w,"s", sep="")
			x[,i] <- sprintf(cmd, x[,i])
		}
	}
	tmp <- rowPaste(x, sep=sep)
	tmp <- c(tmp, "") # to add a newline at end
	cat(paste(tmp, collapse="\n"))
}



#########################################################
#
# Install necessary code packages & annot packages.
# This function should be used if the tar.gz or .zip bundles are sitting inside libdir.
# If you might need to also download the package from BioC, or load the version that may
# be installed in the system-wide library, then use load.required.annot.packages
#
# Mark Cowley, 2010-03-23
load.required.code.packages <- function(packages, libdir, verbose=FALSE) {
	
	if(libdir!='') {
		setLibPath(libdir)
	}
	debug.message(sprintf("Library path: %s\n", .libPaths()))
	
	if( Sys.info()[["sysname"]] != "Windows" )
		options(pkgType="source") # This may fail on mac's without developer tools installed.

	for(package in packages) {
		debug.message(paste("installing", package, "\t"))
		install.package.localdir(package, libdir, libdir, verbose=verbose)
		debug.message(paste("loading", package, "\n"))
		# check if the package is already 'loaded'. (VERY useful for debugging)
		tmp <- paste("package:", package, sep="")
		if( tmp %in% search() ) detach(tmp, character.only=TRUE)
		if( verbose )
			library(package, character.only=TRUE, verbose=verbose, warn=!verbose, lib.loc=libdir)
		else {
			# Some packages keep spitting warnings when run on a headless server; they should be suppressed
			# 1. qvalue loads tcltk which spits this error: 
			#     "In fun(...) : no DISPLAY variable so Tk is not available"
			if( package %in% c("qvalue")) {
				suppressWarnings(suppressMessages(library(package, character.only=TRUE, verbose=verbose, warn=!verbose, lib.loc=libdir)))
			}
			else {
				suppressMessages(library(package, character.only=TRUE, verbose=verbose, warn=!verbose, lib.loc=libdir))
			}
		}
	}
}

#########################################################
# Function designed specifically for installing annotation packages, which tend to be huge (~250MB for a pd exon)
#
# Rather than include them in the distributable, this does the following:
#	- if the package is already installed in the libdir, then nothing to do
#	- attempts to install the package from a local tar.gz or .zip (windows) package
#	- if package is installed in the main library location (ie the system wide one in /usr/local/R/library for eg), then load that
#	- attempts to download tar.gz or .zip from Bioconductor
#
# I've deliberately avoided .tgz files since most mac's running R will probably have the Xcode dev utils installed also. This can be changed if necessary.
#
# Mark Cowley, 2010-03-23
#
load.required.annot.packages <- function(packages, libdir, verbose=TRUE) {
	res <- rep(FALSE, length(packages)); names(res) <- packages
	
	for( package in packages ) {
		# first 2 lines from oligo:requireAnnotation
		stopifnot(is.character(package), !missing(package))

		# 1. check if package is already installed inside libdir (ie where the genepattern modules lives).	
		status <- is.package.installed(libdir, package)
    
		# 2. try to install from a local package
		if( !status ) {
			status <- install.package.localdir(package, libdir, libdir, verbose=verbose)
		}

		# 3. try the R library branch
		if( !status ) {
			# status <- suppressMessages(require(package, character.only=TRUE, quietly=!verbose, lib=.libPaths()))
			tmp <- sapply(.libPaths(), is.package.installed, package)
			status <- any( tmp )
			if( status )
				libdir <- (.libPaths()[tmp])[1]
		}

		# 4. try to install from Bioconductor
		if( !status ) {
			# try BioC.
			cat("Attempting to download annotation package from Bioconductor.\n")
			status <- pdPkgFromBioC(package, lib = libdir, verbose = verbose) # oligo function
			
			R.css <- file.path(lib.loc, "R.css")
			if( file.exists(R.css) ) unlink(R.css)
		}
	
		# Finally, load the library
		if( status ) {
			if( DEBUG ) {
				library(package, character.only=TRUE, verbose=FALSE, warn=FALSE, lib.loc=libdir)
			}
			else {
				suppressMessages(library(package, character.only=TRUE, verbose=FALSE, warn=FALSE, lib.loc=libdir))
			}
		}
		else {
			cat("Package was not found, and could not be downloaded from BioC.\n")
		}
		
		res[package] <- status
	}
	
	return( res )
}
